@@ -3,16 +3,16 @@ Module that provide non-asynchronous functions for using the somtoday API
 
 '''
 import base64
-from typing import Any, Union
+import re
+from typing import Any, Union, Generator
 from pathlib import Path
 from datetime import datetime
 from dataclasses import dataclass
-import hashlib
-import os
-import random
-import string
+from hashlib import sha256
+from random import choice
+from string import ascii_lowercase, digits
 from urllib.parse import urlparse, parse_qs
-import requests
+import httpx
 
 
 class PasFoto:
@@ -127,7 +127,7 @@ class Student:
         if isinstance(__value, Student):
             return self.name == __value.name and self.school_name == __value.school_name
         return NotImplemented
-    def yield_fetch_cijfers(self, lower_bound_range: int, upper_bound_range: int) -> None:
+    def yield_fetch_cijfers(self, lower_bound_range: int, upper_bound_range: int) -> Generator[Cijfer, Cijfer, Cijfer]:
         """yields the cijfers by calling self.fetch_cijfers() and yielding it results
         Args:
             lower_bound_range (int):  minimum to return (must be greater than 0 and fewer than 100)
@@ -168,7 +168,7 @@ class Student:
         params = {
             "additional": ['berekendRapportCijfer']
         }
-        response = requests.get(f"{self.endpoint}/rest/v1/resultaten/huidigVoorLeerling/{self.indentifier}",
+        response = httpx.get(f"{self.endpoint}/rest/v1/resultaten/huidigVoorLeerling/{self.indentifier}",
                                 params=params, headers=headers)
         if response.status_code >= 200 and response.status_code < 300:
             to_dict = response.json()
@@ -224,51 +224,51 @@ class Student:
             "additional": ["vak", "docentAfkortingen"],
             "sort": "asc-beginDatumTijd"
         }
-        with requests.get(f"{self.endpoint}/rest/v1/afspraken",
+        response = httpx.get(f"{self.endpoint}/rest/v1/afspraken",
                           headers={"Accept": "application/json",
                                    "Authorization": f"Bearer {self.access_token}"},
                           params=params_payload,
-                          timeout=30) as response:
-            as_json = response.json()
-            items: list[dict] = as_json["items"]
-            self.school_subjects = []
-            groups: dict[str, list[Subject]] = {}
-            for item in items:
-                school_object_dict: dict = item.get(
-                    "additionalObjects").get("vak")
-                afkorting = school_object_dict.get("afkorting")
-                subject_name: str = school_object_dict.get("naam")
-                locatie: str = item.get("locatie")
-                begin_lesuur: int = item.get("beginLesuur")
-                eind_lesuur: int = item.get("eindLesuur")
-                docent_afkorting: str = item.get(
-                    "additionalObjects").get("docentAfkortingen")
-                begin_time: datetime = datetime.fromisoformat(
-                    item.get("beginDatumTijd"))
-                end_time: datetime = datetime.fromisoformat(
-                    item.get("eindDatumTijd"))
-                target_object = Subject(subject=subject_name,
-                                        begindt=begin_time,
-                                        enddt=end_time,
-                                        subject_short=afkorting,
-                                        beginhour=begin_lesuur,
-                                        endhour=eind_lesuur,
-                                        location=locatie,
-                                        teacher_shortcut=docent_afkorting)
-                if group_by_day and begin_time.strftime("%Y-%m-%d") in groups:
-                    group = groups.get(
-                        begin_time.strftime("%Y-%m-%d"))
-                    group.append(target_object)
-                elif group_by_day and begin_time.strftime("%Y-%m-%d") not in groups:
-                    new_group = {
-                        begin_time.strftime("%Y-%m-%d"): [target_object]
-                    }
-                    groups.update(new_group)
-                else:
-                    self.school_subjects.append(target_object)
-            if group_by_day:
-                self.school_subjects = [groups.get(x) for x in groups]
-            return self.school_subjects
+                          timeout=30) 
+        as_json = response.json()
+        items: list[dict] = as_json["items"]
+        self.school_subjects = []
+        groups: dict[str, list[Subject]] = {}
+        for item in items:
+            school_object_dict: dict = item.get(
+                "additionalObjects").get("vak")
+            afkorting = school_object_dict.get("afkorting")
+            subject_name: str = school_object_dict.get("naam")
+            locatie: str = item.get("locatie")
+            begin_lesuur: int = item.get("beginLesuur")
+            eind_lesuur: int = item.get("eindLesuur")
+            docent_afkorting: str = item.get(
+                "additionalObjects").get("docentAfkortingen")
+            begin_time: datetime = datetime.fromisoformat(
+                item.get("beginDatumTijd"))
+            end_time: datetime = datetime.fromisoformat(
+                item.get("eindDatumTijd"))
+            target_object = Subject(subject=subject_name,
+                                    begindt=begin_time,
+                                    enddt=end_time,
+                                    subject_short=afkorting,
+                                    beginhour=begin_lesuur,
+                                    endhour=eind_lesuur,
+                                    location=locatie,
+                                    teacher_shortcut=docent_afkorting)
+            if group_by_day and begin_time.strftime("%Y-%m-%d") in groups:
+                group = groups.get(
+                    begin_time.strftime("%Y-%m-%d"))
+                group.append(target_object)
+            elif group_by_day and begin_time.strftime("%Y-%m-%d") not in groups:
+                new_group = {
+                    begin_time.strftime("%Y-%m-%d"): [target_object]
+                }
+                groups.update(new_group)
+            else:
+                self.school_subjects.append(target_object)
+        if group_by_day:
+            self.school_subjects = [groups.get(x) for x in groups]
+        return self.school_subjects
 
     def __repr__(self):
         return f"{self.full_name}, {self.school_name}"
@@ -280,30 +280,30 @@ class Student:
         """description: generates data(not meant to be called)
 
         Returns:
-            bool: if it fetched and loaded data by success.
+            bool: if it fetched and loaded data with success.
         """
         if hasattr(self, "full_name"):
             return False
         else:
-            with requests.get(f"{self.endpoint}/rest/v1/leerlingen",
+            name_response = httpx.get(f"{self.endpoint}/rest/v1/leerlingen",
                               headers={
                                   "Authorization":
                                       f"Bearer {self.access_token}",
                                   "Accept": "application/json"},
                               params={"additional": ["pasfoto", "leerlingen"]},
 
-                              timeout=30) as name_response:
-                to_dict = name_response.json()["items"][0]
-                self.pasfoto = PasFoto(to_dict["additionalObjects"]["pasfoto"]["datauri"])
-                self.full_name = to_dict.get(
-                    "roepnaam") + " " + to_dict.get("achternaam")
-                self.indentifier = to_dict.get("links")[0]["id"]
-                self.email = to_dict.get("email")
-                self.leerlingnummer = to_dict.get("leerlingnummer")
-                self.gender = "Male" if to_dict.get(
-                    "geslacht") == "Man" else "Female"
-                year, month, day = to_dict.get("geboortedatum").split("-")
-                self.birth_datetime = datetime(int(year), int(month), int(day))
+                              timeout=30) 
+            to_dict = name_response.json()["items"][0]
+            self.pasfoto = PasFoto(to_dict["additionalObjects"]["pasfoto"]["datauri"])
+            self.full_name = to_dict.get(
+                "roepnaam") + " " + to_dict.get("achternaam")
+            self.indentifier = to_dict.get("links")[0]["id"]
+            self.email = to_dict.get("email")
+            self.leerlingnummer = to_dict.get("leerlingnummer")
+            self.gender = "Male" if to_dict.get(
+                "geslacht") == "Man" else "Female"
+            year, month, day = to_dict.get("geboortedatum").split("-")
+            self.birth_datetime = datetime(int(year), int(month), int(day))
         return True
 
     @property
@@ -325,7 +325,6 @@ class School:
     def __init__(self, school_name: str, uuid: str):
         self.school_name = school_name
         self.school_uuid = uuid
-        self.__failed_time = 0
 
     @staticmethod
     def from_school_uuid(uuid: str) -> "School":
@@ -339,16 +338,16 @@ class School:
         Returns:
             School: The school what it found
         """
-        with requests.get("https://servers.somtoday.nl/organisaties.json",
-                          timeout=30) as school_response:
-            as_dict = school_response.json()
-            instellingen = as_dict[0]["instellingen"]
-            exists = tuple(filter(lambda school_: school_[
-                                                      "uuid"] == uuid, instellingen))
-            if exists:
-                return School(exists[0]["naam"], uuid)
-            else:
-                raise ValueError(f"Invalid  uuid: {uuid}")
+        school_response = httpx.get("https://servers.somtoday.nl/organisaties.json",
+                          timeout=30)
+        as_dict = school_response.json()
+        instellingen = as_dict[0]["instellingen"]
+        exists = tuple(filter(lambda school_: school_[
+                                                    "uuid"] == uuid, instellingen))
+        if exists:
+            return School(exists[0]["naam"], uuid)
+        else:
+            raise ValueError(f"Invalid  uuid: {uuid}")
 
     @staticmethod
     def from_school_name(name: str) -> "School":
@@ -360,7 +359,15 @@ class School:
             School: The School self
         """
         return find_school(name)
+    
+    @staticmethod
+    def _generate_random_str(length: int) -> str:
+        return "".join([choice(ascii_lowercase + digits) for _ in range(length)])
 
+    
+    @staticmethod
+    def parse_query_url(key: str, url: str) -> Union[str, None]:
+        return parse_qs(urlparse(url).query)[key]
     def get_student(self, name: str, password: str) -> Student:
         """description: Gets the student by name and password(without SSO)
         Args:
@@ -369,124 +376,106 @@ class School:
 
         Raises:
             ValueError: Credentials are incorrect.
-            requests.exceptions.RequestException: Error at  last https request. Rarely happens.
+            httpx.exceptions.RequestException: Error at  last https request. Rarely happens.
 
         Returns:
             Student: The student self.
         """
-        if self.__failed_time > 10:
-            raise ValueError("Credentials might be invalid. Please check")
-        cookies_saved = []
-        token = base64.urlsafe_b64encode(os.urandom(32)) \
-            .rstrip(b'=') \
-            .decode()
-        hashed = base64.urlsafe_b64encode(hashlib.sha256(token.encode()).digest()) \
-            .rstrip(b'=') \
-            .decode()
-        payload = {
-            "response_type": "code",
-            "redirect_uri": "somtodayleerling://oauth/callback",
-            "code_challenge": hashed,
-            "tenant_uuid": self.school_uuid,
-            "code_challenge_method": "S256",
-            "state": "".join(random.choices(string.ascii_letters, k=20)),
-            "scope": "openid",
-            "client_id": "D50E0C06-32D1-4B41-A137-A9A850C892C2",
-            "session": "no_session"
 
-        }
-        response = requests.get(
-            "https://inloggen.somtoday.nl/oauth2/authorize",
-            params=payload,
-            allow_redirects=False,
-            timeout=30)
-        cookies_saved = response.cookies
-        parsed = urlparse(response.headers.get("location"))
-        auth_token = parse_qs(parsed.query).get("auth")[0]
+        with httpx.Client(follow_redirects=False) as session:
+            codeVerifier: str = self._generate_random_str(128)
+            codeChallenge = base64.b64encode(sha256(codeVerifier.encode()).digest()).decode()
+            codeChallenge = re.sub(r"\+", "-", codeChallenge)
+            codeChallenge = re.sub(r"\/", "_", codeChallenge)
+            codeChallenge = re.sub(r"=+$", "", codeChallenge)
 
-        post_headers = {
-            "content-type": "application/x-www-form-urlencoded",
-            "origin": "https://inloggen.somtoday.nl"
-        }
-        firstpos = requests.post("https://inloggen.somtoday.nl/", data={
-            "loginLink": "x",
-            "usernameFieldPanel:usernameFieldPanel_body:usernameField": name
-        }, params={"-1.-panel-signInForm": "",
-                   "auth": auth_token},
-                                 headers=post_headers, allow_redirects=False, cookies=cookies_saved, timeout=30)
-
-        cookies_saved = firstpos.cookies
-        secondpos = requests.post(
-            "https://inloggen.somtoday.nl/login",
-            data={
-                "loginLink": "x",
-                "passwordFieldPanel:passwordFieldPanel_body:passwordField": password
-            },
-            headers=post_headers,
-            params={
-                "1-1.-passwordForm": "",
-                "auth": auth_token
-            },
-            cookies=cookies_saved,
-            allow_redirects=False,
-            timeout=30
-        )
-        location2 = secondpos.headers.get('location')
-        code_as_return: str = ''
-        if location2.startswith("somtodayleerling:"):
-            parsed_url = urlparse(location2)
-            code_as_return: str = parse_qs(parsed_url.query).get("code")[0]
-        else:
-            thirdpost = requests.post(
-                "https://inloggen.somtoday.nl/login",
-                data={
-                    "loginLink": "x",
-                    "passwordFieldPanel:passwordFieldPanel_body:passwordField": password
-                },
-                headers=post_headers,
+            response = session.get(
+                "https://inloggen.somtoday.nl/oauth2/authorize",
                 params={
-                    "1-1.-passwordForm": "",
-                    "auth": auth_token
+                    "redirect_uri": "somtodayleerling://oauth/callback",
+                    "client_id": "D50E0C06-32D1-4B41-A137-A9A850C892C2",
+                    "state": self._generate_random_str(8),
+                    "response_type": "code",
+                    "scope": "openid",
+                    "tenant_uuid": self.school_uuid,
+                    "session": "no_session",
+                    "code_challenge": codeChallenge,  # TODO
+                    "code_challenge_method": "S256",
                 },
-                cookies=cookies_saved,
-                allow_redirects=False,
-                timeout=30
             )
-            location3 = thirdpost.headers.get("location")
-            if location3.startswith("somtodayleerling:"):
-                parsed_url = urlparse(location3)
-                code_as_return: str = parse_qs(parsed_url.query).get("code")[0]
+            session.get(response.headers['location'])
+            authorization_code = self.parse_query_url("auth", response.headers['location'])
+            response = session.post(
+                "https://inloggen.somtoday.nl/?0-1.-panel-signInForm",
+                params={"auth": authorization_code},
+                headers={"origin": "https://inloggen.somtoday.nl"},
+            )
+            if 'auth=' in response.headers['Location']: # username + password directly 
+                    data = {
+                        "loginLink": "x",
+                        "usernameFieldPanel:usernameFieldPanel_body:usernameField": name,
+                        "passwordFieldPanel:passwordFieldPanel_body:passwordField": password
+                    }
+                    response = session.post(
+                        "https://inloggen.somtoday.nl/?0-1.-panel-signInForm",
+                        data=data,
+                        headers={"origin": "https://inloggen.somtoday.nl"},
+                        params={"auth": authorization_code},
+                    )
+
+            else: # first username, then password 
+                data = {
+                    'loginLink': 'x',
+                    "passwordFieldPanel:passwordFieldPanel_body:passwordField": password
+                }
+                response = httpx.post(
+                    "https://inloggen.somtoday.nl/login?2-1.-passwordForm",
+                    headers={
+                        "Content-Type": "application/x-www-form-urlencoded",
+                        "Origin": "https://inloggen.somtoday.nl",
+                    },
+                    data=data,
+                )
+            callback_oauth: str = response.headers['Location']
+            if callback_oauth.startswith("somtodayleerling://"):
+                params = {
+                    "grant_type": "authorization_code",
+                    "session": "no_session",
+                    "scope": "openid",
+                    "client_id": "D50E0C06-32D1-4B41-A137-A9A850C892C2",
+                    "tenant_uuid": self.school_uuid,
+                    "code": self.parse_query_url('code', callback_oauth),
+                    "code_verifier": codeVerifier,
+                }
+                response = session.post(
+                    "https://inloggen.somtoday.nl/oauth2/token",
+                    params=params,
+                    headers={"Content-Type": "application/x-www-form-urlencoded"},
+                )
+                response_json = response.json()
+                return Student(name=name,
+                password=password,
+                uuid=self.school_uuid,
+                literal_school=self.school_name,
+                auth_code=authorization_code,
+                access=response_json["access_token"],
+                refresh=response_json["refresh_token"])
             else:
-                self.__failed_time += 1
-                return self.get_student(name, password)
-        last_response_payload = {
-            "grant_type": "authorization_code",
-            "scope": "openid",
-            "client_id": "D50E0C06-32D1-4B41-A137-A9A850C892C2",
-            "tenant_uuid": self.school_uuid,
-            "session": "no_session",
-            "code": code_as_return,
-            "code_verifier": token
-        }
-        last_response_headers = {
-            "content-type": "application/x-www-form-urlencoded"
-        }
-        last_response_final_response = requests.post(
-            "https://inloggen.somtoday.nl/oauth2/token", data=last_response_payload,
-            headers=last_response_headers,
-            timeout=30)
-        if last_response_final_response.status_code == 200:
-            to_dict = last_response_final_response.json()
-            return Student(name=name,
-                           password=password,
-                           uuid=self.school_uuid,
-                           literal_school=self.school_name,
-                           auth_code=auth_token,
-                           access=to_dict["access_token"],
-                           refresh=to_dict["refresh_token"])
-        else:
-            raise requests.exceptions.RequestException(
-                f"request failed. request code {last_response_final_response.status_code}")
+                raise Exception('todo')
+                
+            
+        # if last_response_final_response.status_code == 200:
+        #     to_dict = last_response_final_response.json()
+        #     return Student(name=name,
+        #                    password=password,
+        #                    uuid=self.school_uuid,
+        #                    literal_school=self.school_name,
+        #                    auth_code=auth_token,
+        #                    access=to_dict["access_token"],
+        #                    refresh=to_dict["refresh_token"])
+        # else:
+        #     raise httpx.exceptions.RequestException(
+        #         f"request failed. request code {last_response_final_response.status_code}")
 
 
 def find_school(school_name: str) -> School:
@@ -496,16 +485,17 @@ def find_school(school_name: str) -> School:
         school_name (str): The school's name
 
     Raises:
-        ValueError: The school's name is incorrect.
+        ValueError: The school_name parameter is incorrect.
 
     Returns:
-        School: The school  itself
+        School: A school object representing the school name + school uuid (tenant_uuid)
     """
-    with requests.get("https://servers.somtoday.nl/organisaties.json", timeout=30) as schoolresponse:
-        response_as_dict = schoolresponse.json()
-        final_result = tuple(filter(lambda school_dict: school_dict["naam"].lower(
-        ) == school_name.lower(), response_as_dict[0]["instellingen"]))
-        if final_result:
-            return School(school_name, final_result[0]["uuid"])
-        else:
-            raise ValueError(f"{school_name} does not exist")
+    schoolresponse = httpx.get("https://servers.somtoday.nl/organisaties.json", timeout=30) 
+    response_as_dict = schoolresponse.json()
+    final_result = tuple(filter(lambda school_dict: school_dict["naam"].lower(
+    ) == school_name.lower(), response_as_dict[0]["instellingen"]))
+
+    if final_result:
+        return School(school_name, final_result[0]["uuid"])
+    else:
+        raise ValueError(f"{school_name} does not exist")
